/*
 * Classifier.h  -  Base classifier class
 *
 * Copyright: (C) 2004 by Konstantin Tretyakov
 * Created: March 29, 2004
 *
 *
 *
 *   This program is free software; you can redistribute it 
 *   and/or modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later
 *   version.
 *
 */

#ifndef __Classifier_h__
#define __Classifier_h__

#include "Instance.h"
#include <vector>
#include <ctime>
using namespace std;

/**
 * The base classifier class. Specifies the interface only.
 */
class Classifier
{
public:
    /**
     * The name of the classifier
     */
    virtual string getName() = 0;
    
    /**
     * Train the classifier on data
     */
    virtual void train(const Dataset* data) = 0;

    /**
     * Classify given instance
     */
    virtual MailClass classify(const Instance& inst) const = 0;
};


/**
 * Trivial classifier classifies everything as one same class
 */
class TrivialClassifier: public Classifier
{
private:
    MailClass category;
public:
    // category specifies the class as which this classifier will classify
    // everything.
    TrivialClassifier(MailClass category): category(category) {}
    virtual ~TrivialClassifier(){};

    string getName()
    {
        return "Trivial Classifier";
    }
    
    void train(const Dataset* data)
    { // Ignore
    }

    MailClass classify(const Instance& inst) const
    {
        return category;
    }
};


/**
 * Utility class used to estimate classifier performance
 */
class ClassifierPerformanceEstimator
{
public:
    static void testClassifier(Classifier* c, const Dataset* trainingSet, const Dataset* validationSet)
    {
        cout << "Training classifier: " << c->getName() << endl;
        clock_t trainStart = clock();
        c->train(trainingSet);
        clock_t trainEnd = clock();
        cout << "Testing classifier performance" << endl;
        
        int fp = 0;
        int fn = 0;
        int num = validationSet->instances.size();

        clock_t validStart = clock();
        for (vector<Instance>::const_iterator i = validationSet->instances.begin();
             i != validationSet->instances.end(); i++)
        {
            MailClass real = i->category;
            MailClass predicted = c->classify(*i);
            if (real != predicted)
            {
                if (real == LEGITIMATE)
                {
                    fp++;
                    cout << "+" << flush;
                }
                else
                {
                    fn++;
                    cout << "!" << flush;
                }
            }
            else cout << "." << flush;
        }
        clock_t validEnd = clock();
        cout << endl;
        cout << "--------------------------------------------------------------" << endl;
        cout << "  Results (" << c->getName() << "): " << endl;
        cout << "  Validation set size: " << num << endl;
        cout << "  False positives (legitimate classified as spam): " << fp << endl;
        cout << "  False negatives (spam classified as legitimate): " << fn << endl;
        cout << "  Training time: " << ((double)(trainEnd - trainStart))/CLOCKS_PER_SEC << endl;
        cout << "  Validation time: " << ((double)(validEnd - validStart))/CLOCKS_PER_SEC << endl;
        cout << "--------------------------------------------------------------" << endl;
    }
};

#endif

