/*
 * Configuration.h - a class for parsing conf files.
 *
 * Copyright: (C) 2004 by Konstantin Tretyakov
 * Created: March 30, 2004
 *
 *
 *
 *   This program is free software; you can redistribute it 
 *   and/or modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later
 *   version.
 *
 */

#ifndef __Configuration_h__
#define __Configuration_h__

#include <iostream>
#include <string>
#include <algorithm>
#include <fstream>

#include "Exception.h"
using namespace std;

// A painful way of including hash_map
// See: http://gcc.gnu.org/onlinedocs/libstdc++/faq/index.html#5_4
      #ifdef __GNUC__
      #if __GNUC__ < 3
        #include <hash_map.h>
        namespace Sgi { using ::hash_map; }; // inherit globals
      #else
        #include <ext/hash_map>
        #if __GNUC_MINOR__ == 0
          namespace Sgi = std;               // GCC 3.0
        #else
          namespace Sgi = ::__gnu_cxx;       // GCC 3.1 and later
        #endif
      #endif
      #else      // ...  there are other compilers, right?
        namespace Sgi = std;
      #endif
using Sgi::hash_map;
using Sgi::hash;

// String comparison predicate for hashmap
struct eqStr
{
    bool operator()(const string& lhs, const string& rhs) const
    {
        return lhs == rhs;
    }
};

// String hash
struct hashStr
{
    hash<const char*> h;
    bool operator ()(const string& s) const
    {
        return h(s.c_str());
    }
};
    
/**
 * Helper class to read configuration data
 */
class Configuration
{
public:
    hash_map<const string, string, hashStr, eqStr> params;    

    /**
     * Constructor: takes the name of the config file and the command line
     * and parses these filling the hashmap "params" with parameters.
     * On failure throws Exception
     */
    Configuration(const string& filename, int argc, const char* const argv[]):
        params()
    {
      readConfFile(filename);
      readCommandLine(argc, argv);
    }

    /**
     * Read parameters from the conf file.
     * If a parameter read from file is already present in the
     * "params" map, it gets overwritten.
     */
    void readConfFile(const string& filename)
    {
        ifstream fi(filename.c_str());
        if (!fi) throw Exception("Cannot read file " + string(filename));
        char line[500];

        // NB! Bad code!
        while (fi)
        {
            // Read line
            fi.getline(line, 500);

            string sLine(line);
            istringstream isLine(sLine);

            // Read parameter name
            string paramName;
            isLine >> paramName;

            if (paramName.length() == 0) continue;
            if (paramName[0] == '#') continue;

            // Skip whitespace
            while (isLine && (isLine.peek() == ' ' || isLine.peek() == '\t')) isLine.ignore(1);

            // The rest of the line is parameter value
            char pValue[500];
            isLine.getline(pValue, 500);
            string paramValue(pValue);

            params[paramName] = paramValue;
        }
        fi.close();
    }
    
    /**
     * Read parameters from the command line.
     * If a parameter is given via command line is already present
     * in the "params" map, it is overwritten.
     * No "smart" parsing is done. It is assumed that each
     * parameter is specified as --param=value with no spaces inbetween.
     */
    void readCommandLine(int argc, const char* const argv[])
    {
        try
        {
            for (int i = 1; i < argc; i++)
            {
                string param(argv[i]);
                if (param.length() < 5) throw Exception();
                if (param[0] != '-' || param[1] != '-') throw Exception();

                // Find the separating '='
                int j = 2;
                while (j < (int)param.length() && param[j] != '=') j++;
                if (j == (int)param.length()) throw Exception();

                string paramName = param.substr(2, j - 2);
                string paramValue = param.substr(j+1);
                params[paramName] = paramValue;
            }
        }
        catch (Exception e)
        {
            throw Exception("Expecting command line parameters in format --param=value");
        }
    }

    /**
     * This procedure sets the given parameter to a given value,
     * IF it has not yet been set.
     * This is to be used to set the default parameter values
     * after the configuration has been read in from file/command line.
     */
    void setDefault(const string& paramName, const string& paramValue)
    {
        if (params.find(paramName) == params.end()) params[paramName] = paramValue;
    }

    /**
     * Returns an int parameter checking that its value is between
     * given bounds. On failure throws Exception.
     */
    int getIntParam(const string& paramName, int minValue, int maxValue) const
    {
        if (params.find(paramName) == params.end()) throw Exception("Parameter missing: " + paramName);
        int result = atoi(params.find(paramName)->second.c_str());
        if (result < minValue || result > maxValue)
            throw Exception("Parameter " + paramName + " must be in range " + minValue + ".." + maxValue);
        return result;
    }

    /**
     * Parses the parameter that consists of integers separated with commas
     * returns a vector of integers.
     */
    vector<int> getIntVectorParam(const string& paramName, int minValue, int maxValue) const
    {
        if (params.find(paramName) == params.end()) throw Exception("Parameter missing: " + paramName);
        string s = params.find(paramName)->second;
        istringstream sin(s);
        vector<int> result;
        int v;
        char c;
        sin >> v;
        if (v < minValue || v > maxValue)
            throw Exception("Parameter " + paramName + " must be a list of values in range " + minValue + ".." + maxValue);        
        result.push_back(v);
        while (sin && (sin.peek() == ' ' || sin.peek() == '\t')) sin >> c;
           
        while(sin)
        {
            if (sin && (sin.peek() != ',')) throw Exception("Parameter " + paramName + " format error: a comma-separated list of numbers expected");
            sin >> c;
            sin >> v;
            if (v < minValue || v > maxValue)
                throw Exception("Parameter " + paramName + " must be a list of values in range " + minValue + ".." + maxValue);
            result.push_back(v);
            while (sin && (sin.peek() == ' ' || sin.peek() == '\t')) sin >> c;
        }

        return result;
    }
        
};

#endif

