/*
 * FeatureExtractor.h  -  Classes for extracting features from strings
 *
 * Copyright: (C) 2004 by Konstantin Tretyakov
 * Created: March 29, 2004
 *
 *
 *
 *   This program is free software; you can redistribute it 
 *   and/or modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later
 *   version.
 *
 */


#ifndef __FeatureExtractor_h__
#define __FeatureExtractor_h__

#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <algorithm>

#include "Instance.h"
#include "FileUtil.h"
#include "Exception.h"
using namespace std;


/**
 * A feature extractor, i.e. a class containing a single
 * function that converts a string to an Instance
 */
class FeatureExtractor
{
public:
    // Given string s return feature vector corresponding to it
    virtual Instance makeInstance(const string& s) const = 0;
};

/**
 * A specific feature extractor suitable
 * for the PU1 spam corpus
 */
class Pu1FeatureExtractor: public FeatureExtractor
{
public:
    int startWord;
    int endWord;

    // Constructor.
    // startWord - the word corresponding to the first attribute
    // endWord - the word corresponding to the last attribute
    Pu1FeatureExtractor(int startWord, int endWord):
        startWord(startWord), endWord(endWord)
    {}

    virtual ~Pu1FeatureExtractor(){}

    virtual Instance makeInstance(const string& s) const
    {
        Instance result(endWord - startWord + 1);
        istringstream in(s);

        // Skip the "Subject:" text
        char c = ' ';
        while (c != ':') in >> c;

        // Now read the ints from string, for
        // those less than numAttrib set corresponding
        // attributes in the instance
        int n;
        while (in)
        {
            in >> n;
            if (n >= startWord && n <= endWord)
                result.setAttribute(n - startWord + 1, true);
        }
        return result;
    }
};


/**
 * A class used to convert messages in the PU1 spam
 * corpus to Instance objects
 */
class Pu1InstanceReader
{
public:
    // The root directory containing the spam corpus
    string rootDir;
    const FeatureExtractor* f;

    // Create a new object
    // f is a feature extractor object
    Pu1InstanceReader(const string& rootDir, const FeatureExtractor* f)
    {
        if (rootDir[rootDir.length() - 1] != '/') this->rootDir = rootDir + '/';
        else this->rootDir = rootDir;
        this->f = f;
    }

    // Reads an instance object from file.
    // File path is given relative to rootDir
    // Throws Exception on failure.
    Instance readInstanceFromFile(const string& filename) const
    {
        string file = rootDir + filename;
        string s = readFile(file);

        if (!f) throw Exception("Feature extractor not specified");
        Instance inst = f->makeInstance(s);

        // Set the category of the instance. It is legitimate if filename
        // contains "legit", spam if filename contains "spm"
        if (filename.find("legit", 0) != string::npos) inst.category = LEGITIMATE;
        else if (filename.find("spm", 0) != string::npos) inst.category = SPAM;
        return inst;
    }

    // Reads all files in directory given as relative path
    // to rootDir
    vector<Instance> readInstancesFromDirectory(const string& dir) const
    {
        string fulldir = rootDir + dir + "/";
        vector<Instance> result;
        vector<string> files = listDir(fulldir);
        for (vector<string>::iterator i = files.begin(); i != files.end(); i++)
        {
            result.push_back(readInstanceFromFile(dir + "/" + *i));
        }
        return result;
    }

    // Creates a training set from the PU1 corpus
    // n is a vector of 1..10 - the parts to be left out of training
    Dataset* makePu1TrainingSet(const vector<int>& n)
    {
        vector<int> dirsToAdd; // We'll add directories NOT present in the vector given
        for (int i = 1; i <= 10; i++)
        {
            if (find(n.begin(), n.end(), i) == n.end()) dirsToAdd.push_back(i);
        }

        return makePu1ValidationSet(dirsToAdd);
    }

    // Creates a validation set from the PU1 corpus
    // n is a vector of 1..10 - the parts to take into validation
    Dataset* makePu1ValidationSet(const vector<int>& n)
    {
        Dataset* result = new Dataset();
        for (vector<int>::const_iterator i = n.begin(); i != n.end(); i++)
        {
            if (*i > 10 || *i < 1) throw Exception("1 <= n <= 10 expected");        
            ostringstream os;
            os << "part" << *i;
            result->addInstances(readInstancesFromDirectory(os.str()));
            cout << " added instances from " << os.str() << ", dataset size now " << result->instances.size() << endl;
        }
        return result;
    }
              
};

#endif
