/*
 * Instance.h  -  Base classes for spam categorization
 *
 * Copyright: (C) 2004 by Konstantin Tretyakov
 * Created: March 28, 2004
 *
 *
 *
 *   This program is free software; you can redistribute it 
 *   and/or modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later
 *   version.
 *
 */

#ifndef __Instance_h__
#define __Instance_h__

#include <vector>
#include <iostream>
#include <algorithm>
#include <iterator>
#include <map>

#include "Exception.h"
using namespace std;

/**
 * Represents the category of mail message: Spam/Non-Spam/Undefined
 */
enum MailClass
{
    SPAM, LEGITIMATE, UNDEFINED
};

ostream& operator<<(ostream& out, const MailClass& cat)
{
    if (cat == SPAM) return out << "S";
    else if (cat == LEGITIMATE) return out << "L";
    else return out << "?";
}

    
/**
 * Represents an instance to be classified.
 * Stores the feature vector for the instance.
 */
class Instance
{
public:
    typedef map<int, bool> Attributes;  // An attribute is a pair index->value.
                                        // First attribute has index 1 !
    Attributes attributes;
    MailClass category;
    int attributeCount;  // Number of instance attributes (i.e. the index of the last possible attribute)
                         // NOT the same as attributes.size() because only nonzero attributes are stored

    // Create an instance with numAttrib attributes
    Instance(int attributeCount):
        attributes(),
        category(UNDEFINED),
        attributeCount(attributeCount)
    {}

    // Returns the value of attribute with given index
    // Note that it is not good to used attributes[index] because
    // that will insert values into map
    inline bool getAttribute(int index) const
    {
        if (attributes.find(index) == attributes.end()) return false;
        else return (attributes.find(index)->second);
    }

    inline void setAttribute(int index, bool value)
    {
        if (index > attributeCount || index < 1) throw Exception("Invalid attribute index");
        if (value) attributes[index] = true;
        else if (attributes.find(index) != attributes.end())
            attributes.erase(attributes.find(index));
    }           

    // Output attribute to stream
    friend ostream& operator<<(ostream& out, const Attributes::value_type& attr)
    {
        return out << attr.first << ":" << attr.second;
    }
    
    // Output instance to stream
    friend ostream& operator<<(ostream& out, const Instance& inst)
    {
        out << "Instance(";
        int done = 0;
        for (Attributes::const_iterator i = inst.attributes.begin(); i != inst.attributes.end(); i++)
        {
            if (done++ != 0) out << ", ";
            out << *i;
        }
        out << "; " << inst.category << ")";
        return out;
    }
};


/**
 * Represents a set of instances.
 * Basically an abstraction of vector<Instance>.
 * Dataset must also check that all instances have the same number of
 * attributes (not implemented).
 * Another important difference is that vector<Instance> is usually
 * handled "by value" and copied back and forth, whereas
 * this object is to be used as a "normal class" i.e. by reference.
 */
class Dataset
{
public:
    typedef vector<Instance> Instances;
    Instances instances;
    
    Dataset(): instances() {};
        
    Dataset(Instances instances): instances(instances) {};

    void addInstance(Instance instance)
    {
        instances.push_back(instance);
    }

    void addInstances(vector<Instance> newInstances)
    {
        instances.insert(instances.end(), newInstances.begin(), newInstances.end());
    }
};
   



#endif
