/*
 * KNearestNeighborsClassifier.h  -  Definition of a kNN classifier
 *
 * Copyright: (C) 2004 by Konstantin Tretyakov
 * Created: March 29, 2004
 *
 *
 *
 *   This program is free software; you can redistribute it 
 *   and/or modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later
 *   version.
 *
 */

#ifndef __KNearestNeighborsClassifier_h__
#define __KNearestNeighborsClassifier_h__

#include "Instance.h"
#include "Classifier.h"
#include "Exception.h"
#include <vector>
using namespace std;

/**
 * Distance metric between instances
 */
class Distance
{
public:
    virtual double distance(const Instance& a, const Instance& b) const = 0;
};

/**
 * 1-norm distance
 */
class AbsDistance: public Distance
{
public:
    AbsDistance(){}
    virtual ~AbsDistance(){}
    
    double distance(const Instance& a, const Instance& b) const
    {
        double result = 0;
        if (a.attributeCount != b.attributeCount)
            throw Exception("Cannot compute distance between vectors of different length");

        Instance::Attributes::const_iterator aAttr = a.attributes.begin();
        Instance::Attributes::const_iterator bAttr = b.attributes.begin();

        while (aAttr != a.attributes.end() || bAttr != b.attributes.end())
        {
            if (aAttr == a.attributes.end())
            {
                //result += abs((double)bAttr->second);
                result++;
                bAttr++;
            }
            else if (bAttr == b.attributes.end())
            {
                //result += abs((double)aAttr->second);
                result++;
                aAttr++;
            }
            else if (aAttr->first < bAttr->first)
            {
                //result += abs((double)aAttr->second);
                result++;
                aAttr++;
            }
            else if (aAttr->first > bAttr->first)
            {
                //result += abs((double)bAttr->second);
                result++;
                bAttr++;
            }
            else
            {
                //result += abs((double)aAttr->second - (double)bAttr->second);
                if (aAttr->second != bAttr->second) result++;
                aAttr++;
                bAttr++;
            }
        }
        return result;
    }                   
};


/**
 * K-Nearest Neighbor Classifier classifies instance according to
 * the categories of the instance's K nearest neighbors.
 */
class KNearestNeighborsClassifier: public Classifier
{
private:
    int k;  // Must be positive and odd
    int l;
    Dataset::Instances instances;
    Distance* d;

    class PairSecondArgLess
    {
    public:
        bool operator ()(const pair<int, double> a, const pair<int, double> b)
        {
            return (a.second < b.second);
        }
    };
    
public:
    // Create classifier
    // d - distance metric
    // k - the "K" in "K-nearest-neighbor"
    // l - how many neighbors must be spam in order to
    //     classify mail as spam (default is ceil(k/2)) 
    KNearestNeighborsClassifier(Distance* d, int k, int l)
    {
        if (k < 1 || l > k) throw Exception("K must be positive and not less than L");
        this->k = k;
        this->d = d;
        this->l = l;
    }
    KNearestNeighborsClassifier(Distance* d, int k)
    {
        if (k < 1 || k % 2 == 0) throw Exception("K must be positive and odd");
        this->k = k;
        this->d = d;
        this->l = (k - 1)/2 + 1;
    }
        
    virtual ~KNearestNeighborsClassifier(){};

    string getName()
    {
        return "K Nearest Neighbors Classifier";
    }

    void train(const Dataset* data)
    {
        instances = data->instances;

        // Check that there are no unclassified instances
        for (Dataset::Instances::iterator i = instances.begin(); i != instances.end(); i++)
        {
            if (i->category == UNDEFINED) throw Exception("Undefined instances not allowed in training set");
        }
    }

    MailClass classify(const Instance& inst) const
    {
        if (d == NULL) throw Exception("Distance metric not specified");
        if ((int)instances.size() < k)
            throw Exception("K-Nearest-Neighbor assumes that training set size is at least K");

        // For every instance in the stored set of instances we
        // shall calculate the distance to the instance to be classified
        vector<pair<int, double> > v(instances.size());

        for (int i = 0; i < (int)instances.size(); i++)
        {
            v[i] = pair<int, double>(i, d->distance(inst, instances[i]));
        }

        // Get k nearest instances
        partial_sort(v.begin(), v.begin() + k, v.end(), PairSecondArgLess());

        // See what class dominates
        int numSpam = 0;
        for (vector<pair<int, double> >::iterator i = v.begin(); i != v.begin() + k; i++)
        {
            if (instances[i->first].category == SPAM) numSpam++;
        }

        if (numSpam >= l) return SPAM;
        else return LEGITIMATE;
    }
};

#endif

