/*
 * NaiveBayesClassifier.h  -  Definition of the Naive Bayesian Classifier
 *
 * Copyright: (C) 2004 by Konstantin Tretyakov
 * Created: March 29, 2004
 *
 *
 *
 *   This program is free software; you can redistribute it 
 *   and/or modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later
 *   version.
 *
 */


#ifndef __NaiveBayesClassifier_h__
#define __NaiveBayesClassifier_h__

#include "Instance.h"
#include "Classifier.h"
#include "Exception.h"
#include <vector>
using namespace std;

/**
 * The Naive Bayes classifier
 */
class NaiveBayesClassifier: public Classifier
{
private:
    int numAttrib; // The number of attributes in instances

    struct doublepair
    {
        double value[2];
        double& operator [](int index)
        {
            return value[index];
        }
        double operator[](int index) const
        {
            return value[index];
        }

    };

    vector<doublepair> likelihood; // Store the likelihood ratios for all values of all attributes
    double lambda; // The classification bias
    double pLdivpS; // P(L) / P(S)
    static const double VERY_LARGE_VALUE = 1000000; // Thing we get when divide by 0
    static const double VERY_SMALL_VALUE = 0.0000001; // As we use approximations of infinity we have to use
                                               // approximations of 0 too. So that

public:
    NaiveBayesClassifier(double lambda): lambda(lambda){};
    virtual ~NaiveBayesClassifier(){};

    string getName()
    {
        return "Naive Bayes Classifier";
    }

    void train(const Dataset* data)
    {
        // Reject empty datasets
        int numInstances = data->instances.size();
        if (numInstances == 0) throw Exception("Empty dataset");

        // Calculate P(L) and P(S)
        int numLegit = 0;
        int numSpam = 0;

        Dataset::Instances::const_iterator i;
        for (i = data->instances.begin(); i != data->instances.end(); i++)
        {
            if (i->category == SPAM) numSpam++;
            else if (i->category == LEGITIMATE) numLegit++;
            else throw Exception("Unclassified message in training set");
        }

        if (numLegit == 0 || numSpam == 0) throw
            Exception("Expecting at least one legitimate and one spam message in training set");

        pLdivpS = (double)numLegit/(double)numSpam;


        // Calculate likelihood ratios for all attributes for all values
        numAttrib = data->instances[0].attributeCount;
        if (numAttrib == 0) throw Exception("Instance has no attributes");
        likelihood = vector<doublepair>(numAttrib);

        for (int attrib = 0; attrib < numAttrib; attrib++)
        {
            int numOfOnesInSpam = 0;
            int numOfOnesInLegit = 0;
            for (i = data->instances.begin(); i != data->instances.end(); i++)
            {
                if (i->getAttribute(attrib + 1) == true)
                {
                    if (i->category == SPAM) numOfOnesInSpam++;
                    else numOfOnesInLegit++;
                }
            }

            // Likelihood(x) = P(x | S)/P(x | L)
            if (numOfOnesInSpam == 0) likelihood[attrib][1] = VERY_SMALL_VALUE;
            else if (numOfOnesInLegit == 0) likelihood[attrib][1] = VERY_LARGE_VALUE;
            else likelihood[attrib][1] = ((double)numOfOnesInSpam/(double)numOfOnesInLegit)*pLdivpS;

            int numOfZeroesInLegit = numLegit - numOfOnesInLegit;
            int numOfZeroesInSpam = numSpam - numOfOnesInSpam;
            if (numOfZeroesInSpam == 0) likelihood[attrib][0] = VERY_SMALL_VALUE;
            else if (numOfZeroesInLegit == 0) likelihood[attrib][0] = VERY_LARGE_VALUE;
            else likelihood[attrib][0] = ((double)numOfZeroesInSpam/(double)numOfZeroesInLegit)*pLdivpS;

            /* cout << "Processed attribute " << attrib << ": S1=" << numOfOnesInSpam << "/"
                 << numSpam << ", S0=" << numOfZeroesInSpam << "/" << numSpam << ", L1="
                 << numOfOnesInLegit << "/" << numLegit << ", L0=" << numOfZeroesInLegit
                 << "/" << numLegit << ", N=" << numInstances << ", L(0)="
                 << likelihood[attrib][0] << ", L(1)=" << likelihood[attrib][1] << endl;
             */
        }
    }

    MailClass classify(const Instance& inst) const
    {
        // Calculate the likelihood ratio P(x|S)/P(x|L) for the instance.
        // Being naive we assume its the product of likelihood ratios of
        // attributes.
        double totalLikelihood = 1;
        for (int i = 0; i < numAttrib; i++)
        {
            totalLikelihood = totalLikelihood * likelihood[i][inst.getAttribute(i + 1)];
        }

        // Message qualifies as spam if Likelihood > lambda * P(L)/P(S)
        if (totalLikelihood > lambda * pLdivpS) return SPAM;
        else return LEGITIMATE;
    }
};

#endif
