/*
 * PerceptronClassifier.h  -  Rosenblatt's Perceptron
 *
 * Copyright: (C) 2004 by Konstantin Tretyakov
 * Created: Sun Apr 4 2004
 *
 *
 *
 *   This program is free software; you can redistribute it 
 *   and/or modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later
 *   version.
 *
 */

#ifndef __PerceptronClassifier_h__
#define __PerceptronClassifier_h__

#include "Instance.h"
#include "Classifier.h"
#include "Exception.h"
#include <vector>
using namespace std;

/**
 * The Perceptron
 */
class PerceptronClassifier: public Classifier
{
private:
    // Perceptron implements the linear decision function sign(wx + b)
    vector<double> w;
    double b;

    int maxIterations; // In case the data is not linearly separable
                       // the algorithm will not converge.
                       // This puts a bound on the number of iterations.
public:

    PerceptronClassifier(): maxIterations(10000){};
    
    PerceptronClassifier(int maxIterations):
      maxIterations(maxIterations){};
      
    virtual ~PerceptronClassifier(){};

    string getName()
    {
        return "Rosenblatt's Perceptron";
    }

    void train(const Dataset* data)
    {
        if (data->instances.size() == 0) throw Exception("Empty dataset");
        bool converged = false;
        int iterationsLeft = maxIterations;

        // Initialize the weight vector
        w = vector<double>(data->instances[0].attributeCount);
        b = 0;
        
        while (iterationsLeft-- > 0 && !converged)
        {
            converged = true;

            // Look for instances that are not classified correctly
            // update the weight vector accordingly
            for (Dataset::Instances::const_iterator i = data->instances.begin();
                 i != data->instances.end(); i++)
            {
                if (classify(*i) != i->category)
                {
                    converged = false;
                    // Spam is category +1;
                    // Legitimate = -1;
                    double c = i->category == SPAM ? 1 : -1;

                    b += c;
                    
                    // w += c*x
                    for (Instance::Attributes::const_iterator attr = i->attributes.begin();
                         attr != i->attributes.end();
                         attr++)
                    {
                        int attIndex = attr->first;
                        // w[attIndex - 1] += c*attr->second
                        w[attIndex - 1] += c;
                    }
                }
            }   
        }

        if (iterationsLeft <= 0) throw Exception("Perceptron didn't converge. Most probably data is not linearly separable");
    }

    MailClass classify(const Instance& inst) const
    {
        double result = 0;
        for (Instance::Attributes::const_iterator i = inst.attributes.begin();
             i != inst.attributes.end(); i++)
        {
            int attIndex = i->first;
            // result += w[attIndex - 1]*i->second
            result += w[attIndex - 1];
        }
        if (result + b > 0) return SPAM;
        else return LEGITIMATE;
    }
};

#endif
