/*
 * SVMClassifier.h  -  SVM classifier (adapter class for SVMLight).
 *
 * Copyright: (C) 2004 by Konstantin Tretyakov
 * Created: March 31 2004
 *
 *
 *
 *   This program is free software; you can redistribute it 
 *   and/or modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later
 *   version.
 *
 *   NB: This file #includes files from and therefore uses
 *       the SVMLight package that performs all of the
 *       complex SVM-related stuff.
 *       SVMLight is an independent package made by
 *       Thorsten Joachims, and is subject to its own licence.
 *       See LICENCE.txt in the svm_light/ directory.
 *       See also: http://svmlight.joachims.org
 *
 */

#ifndef __SVMClassifier_h__
#define __SVMClassifier_h__

extern "C" {
    #include "svm_light/svm_common.h"
    #include "svm_light/svm_learn.h"
}

#include <iostream>
#include <vector>

#include "Instance.h"
#include "Exception.h"
using namespace std;

/**
 * SVMClassifier performs SVM-based classification.
 * This class is only an adapter to the functions from the
 * SVMLight package, that perform all the job.
 */
class SVMClassifier: public Classifier
{
private:
    // Variables containing data required for SVMLight procedures

    // Learning parameters
    LEARN_PARM learnParm;
    KERNEL_PARM kernelParm;
    long kernelCacheSize;
    
    // Model (valid after successful training)
    MODEL model;
    
    // Converts an Instance to a DOC structure used by SVMLight
    static void instanceToDOC(const Instance& inst, DOC& doc)
    {
        // Allocate the appropriate number of records
        doc.words = (WORD*) malloc(sizeof(WORD) * (inst.attributes.size() + 1));
        if (doc.words == NULL) throw Exception("Could not allocate memory");

        int curWordIndex = 0;
        for (Instance::Attributes::const_iterator i = inst.attributes.begin(); i != inst.attributes.end(); i++)
        {
            doc.words[curWordIndex].wnum = i->first;
            doc.words[curWordIndex].weight = (FVAL)i->second;
            curWordIndex++;
        }
        doc.words[curWordIndex].wnum = 0;
        doc.words[curWordIndex].weight = 0.0;
        
        doc.queryid = 0;    // Some thing needed for ranking and not needed for classification
        doc.costfactor = 1; // Const of misclassification for this instance.
        doc.docnum=-1;      // Don't know the exact meaning of this, works without it.
        doc.twonorm_sq=sprod_ss(doc.words, doc.words); // Apparently required for inner use by SVMLight.
    }
    
public:
    // Create classifier
    SVMClassifier(bool softmargin = false)
    {
        // Set up default learning parameters

        // Configure learning parameters (default values as in svm_learn_main.c::read_input_parameters)
        learnParm.type = CLASSIFICATION;   // We learn classification
        learnParm.biased_hyperplane = 1;   // Use wx + b decision hyperplane
        learnParm.remove_inconsistent = 0; // If set to yes, examples for which alpha = C will be removed and model retrained
        learnParm.skip_final_opt_check = 0;// Do not skip KT-condition check at end of optimization for examples removed by shrinking
        learnParm.svm_maxqpsize = 10; // "Size of working set". Don't know what's that but 10 should do it
        learnParm.svm_newvarsinqp = 0; // "New variables to enter the working set on new iteration". Let it be at the default value.
        learnParm.svm_iter_to_shrink = -9999; // "Iterations h after which an example can be removed by shrinking".
        learnParm.svm_c = 0.0;        // Upper bound C on alphas
        learnParm.eps  = 1.0;         // Regression epsilon (1.0 for classification)
        learnParm.transduction_posratio = -1.0; // Fraction of unlabeled examples to be classified as positive.
        learnParm.svm_costratio = 1.0; // Factor to multiply C for positive examples
        learnParm.svm_costratio_unlab = 1.0; // ...
        learnParm.svm_unlabbound = 1E-5; // ?
        learnParm.epsilon_crit = 0.001; // Tolerable error for distances used in stopping criterion
        learnParm.epsilon_a = 1E-15; // Tolerable error on alphas at bound
        learnParm.compute_loo = 0;   // Do not compute leave-one-out estimates
        learnParm.rho = 1.0;         // "Parameter in xi/alpha-estimates and for pruning leave-one-out range [1..2]
        learnParm.xa_depth = 0;      // "Parameter in xi/alpha-estimates upper bounding the number of SV the current alpha_t is distributed over
        
        // These are definitely not needed
        learnParm.predfile[0] = 0;
        learnParm.alphafile[0] = 0;

        // Don't know exactly what are these about
        learnParm.epsilon_const = 1E-20;
        learnParm.epsilon_shrink = 1E-6;
        learnParm.opt_precision = 1E-21;

        // Set up the kernel
        kernelParm.custom[0] = 0;
        kernelParm.kernel_type = 0; // 0=linear, 1=poly, 2=rbf, 3=sigmoid, 4=custom
        kernelParm.poly_degree = 3;
        kernelParm.rbf_gamma = 1.0;
        kernelParm.coef_lin = 1; // ?
        kernelParm.coef_const = 1; // ?
        
        kernelCacheSize = 40;

        // Do some consistency verifications to protect me from myself
        if(learnParm.svm_iter_to_shrink == -9999) {
            if(kernelParm.kernel_type == LINEAR)
                learnParm.svm_iter_to_shrink=2;
            else
                learnParm.svm_iter_to_shrink=100;
        }
        if((learnParm.skip_final_opt_check)
            && (kernelParm.kernel_type == LINEAR)) {
            cout << "\nIt does not make sense to skip the final optimality check for linear kernels.\n\n";
            learnParm.skip_final_opt_check=0;
        }
        if((learnParm.skip_final_opt_check)
            && (learnParm.remove_inconsistent))
            throw Exception("It is necessary to do the final optimality check when removing inconsistent examples.");
        if((learnParm.svm_maxqpsize<2))
            throw Exception("Maximum size of QP-subproblems not in valid range [2..].");
        if((learnParm.svm_maxqpsize<learnParm.svm_newvarsinqp))
            throw Exception("Maximum size of QP-subproblems must be larger than the number of new variables entering the working set in each iteration.");
        if(learnParm.svm_iter_to_shrink<1)
            throw Exception("Maximum number of iterations for shrinking not in valid range [1,..]");
        if(learnParm.svm_c<0) 
            throw Exception("The C parameter must be greater than zero!");
        if(learnParm.transduction_posratio>1)
            throw Exception("The fraction of unlabeled examples to classify as positives must be less than 1.0 !!!");
        if(learnParm.svm_costratio<=0)
            throw Exception("The COSTRATIO parameter must be greater than zero!");
        if(learnParm.epsilon_crit<=0) 
            throw Exception("The epsilon parameter must be greater than zero!");
        if(learnParm.rho<0) 
            throw Exception("The parameter rho for xi/alpha-estimates and leave-one-out pruning must be greater than zero (typically 1.0 or 2.0, see T. Joachims, Estimating the Generalization Performance of an SVM Efficiently, ICML, 2000.)!");
        if((learnParm.xa_depth<0) || (learnParm.xa_depth>100))
            throw Exception("The parameter depth for ext. xi/alpha-estimates must be in [0..100] (zero for switching to the conventional xa/estimates described in T. Joachims, Estimating the Generalization Performance of an SVM Efficiently, ICML, 2000.)");
        

        // Set SVMLight verbosity
        verbosity = 1;

        if (!softmargin)
        {
            learnParm.svm_c = 1E15; // Use hard margin classifier
            //learnParm.svm_c = 10000;
        }
        else
        {
            // Some tuning...
            learnParm.svm_costratio = 0.3;
        }
    }

    virtual ~SVMClassifier(){};

    string getName()
    {
        return "Support Vector Machine Classifier (SVMLight-based)";
    }

    void train(const Dataset* data)
    {
        // Check that there are no unclassified instances
        if (data->instances.size() == 0) throw Exception("Nonempty dataset expected on training");
        
        // Translate vector of instances to set of DOCs and labels
        // also initialize some other variables for SVMLight
        DOC *docs;      // Training samples
        double *target; // Labels
        long totwords = data->instances[0].attributeCount; // Number of attributes
        long totdoc = data->instances.size(); // Number of examples

        docs = (DOC*) malloc(sizeof(DOC)*totdoc);
        target = (double *) malloc(sizeof(double)*totdoc);
        if (!docs || !target) throw Exception("Failed to allocate memory");

        int curDoc = 0;
        for (Dataset::Instances::const_iterator i = data->instances.begin(); i != data->instances.end(); i++)
        {
            instanceToDOC(*i, docs[curDoc]);
            if (i->category == UNDEFINED) throw Exception("Undefined instances not allowed in training set");
            else if (i->category == SPAM) target[curDoc] = 1.0;
            else target[curDoc] = -1.0;
            curDoc++;
        }      
        
        
        // Learn the model
        if(kernelParm.kernel_type == LINEAR)
        { // don't need the cache
            cout << "Training classifier" << endl;   
            svm_learn_classification(docs,target,totdoc,totwords,&learnParm,
                                     &kernelParm,NULL,&model);                                      
        }
        else
        {
           // Always get a new kernel cache. It is not possible to use the
           // same cache for two different training runs
           KERNEL_CACHE kernelCache;
           kernel_cache_init(&kernelCache,totdoc,kernelCacheSize);
           svm_learn_classification(docs, target, totdoc, totwords, &learnParm,
                                     &kernelParm, &kernelCache, &model);
           // Free the memory used for the cache.
           kernel_cache_cleanup(&kernelCache);
        }

  /* Warning: The model contains references to the original data 'docs'.
     If you want to free the original data, and only keep the model, you
     have to make a deep copy of 'model'. */
    }

    MailClass classify(const Instance& inst) const
    {
        if (model.sv_num == 0) throw Exception("Untrained classifier!");
        
        DOC doc;
        instanceToDOC(inst, doc);

        if (classify_example(const_cast<MODEL*>(&model), &doc) > 0) return SPAM;
        else return LEGITIMATE;
    }
};

#endif
