/*
 * main.cpp  -  Machine Learning in Spam Classification project main file
 *
 * Copyright: (C) 2004 by Konstantin Tretyakov
 * Created: March 28, 2004
 *
 *
 *
 *   This program is free software; you can redistribute it 
 *   and/or modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later
 *   version.
 *
 */

#include <iostream>

#include "Instance.h"
#include "FeatureExtractor.h"
#include "Classifier.h"
#include "NaiveBayesClassifier.h"
#include "KNearestNeighborsClassifier.h"
#include "SVMClassifier.h"
#include "PerceptronClassifier.h"
#include "OneOfTwoClassifier.h"
#include "TwoOfThreeClassifier.h"
#include "Exception.h"
#include "Configuration.h"
using namespace std;

// Good settings: feature extractor: 10000 features
//                NB: lambda=300 (14 fn)
//                kNN: 15/21     (25 fn)

int main(int argc, char* argv[])
{
  try{
    cout << "Machine Learning in Spam Filtering" << endl;
    cout << "Copyright 2004, Konstantin Tretyakov" << endl << endl;

    // Read configuration
    Configuration config("spam.conf", argc, argv);
    config.setDefault("corpus-directory", "../corpus/pu1_encoded/lemm_bare");
    config.setDefault("corpus-part", "1"); // Which part of the training set will be left out for validation
    config.setDefault("features-from", "70"); // Features will be extracted from this word
    config.setDefault("features-to", "10000");// to this word
    config.setDefault("extended-validation-set", "1"); // Whether to add training instances to the validations set
    config.setDefault("naivebayes-lambda", "1"); // The lambda of the naive bayes
    config.setDefault("knn-k", "1"); // The K in kNN
    config.setDefault("knn-l", "1");
    config.setDefault("svm-softmargin", "0");
    
    config.setDefault("naivebayes", "1");
    config.setDefault("knn", "1");
    config.setDefault("perceptron", "1");
    config.setDefault("trivial", "1");
    config.setDefault("svm", "1");
    config.setDefault("svm-naivebayes", "1");
    config.setDefault("svm-naivebayes-perceptron", "1");
    
    vector<int> corpusPart = config.getIntVectorParam("corpus-part", 1, 10);    
    int featuresFrom = config.getIntParam("features-from", 1, 30000);
    int featuresTo = config.getIntParam("features-to", 1, 30000);
    if (featuresTo < featuresFrom)
        throw Exception("features-to should be greater than features-from");
    int extendedValidation = config.getIntParam("extended-validation-set", 0, 1);
    int nbLambda = config.getIntParam("naivebayes-lambda", 1, 1000);
    int knnK = config.getIntParam("knn-k", 1, 1000);
    int knnL = config.getIntParam("knn-l", 1, 1000);
    if (knnL > knnK)
        throw Exception("knn-k should be greater than knn-l");
    int svmSoftMargin = config.getIntParam("svm-softmargin", 0, 1);
    
    int doNb = config.getIntParam("naivebayes", 0, 1);
    int doKnn = config.getIntParam("knn", 0, 1);
    int doPerceptron = config.getIntParam("perceptron", 0, 1);
    int doTrivial = config.getIntParam("trivial", 0, 1);
    int doSvm = config.getIntParam("svm", 0, 1);
    int doSvmNb = config.getIntParam("svm-naivebayes", 0, 1);
    int doSvmNbPerceptron = config.getIntParam("svm-naivebayes-perceptron", 0, 1);
    

    // Prepare Pu1InstanceReader        
    FeatureExtractor* fe = new Pu1FeatureExtractor(featuresFrom, featuresTo);
    Pu1InstanceReader pu(config.params["corpus-directory"], fe);

    cout << "Preparing training set ( corpus-part=";
    copy(corpusPart.begin(), corpusPart.end(), ostream_iterator<int>(cout, " "));
    cout << ")" << endl;
    Dataset* trainingSet = pu.makePu1TrainingSet(corpusPart);
    cout << "Preparing validation set" << endl;
    Dataset* validationSet = pu.makePu1ValidationSet(corpusPart);
    if (extendedValidation)
        validationSet->addInstances(trainingSet->instances);

    
    // kNN classifier
    Distance* d = new AbsDistance();
    Classifier* knn = new KNearestNeighborsClassifier(d, knnK, knnL);
    // NB classifier
    Classifier* nb = new NaiveBayesClassifier(nbLambda);
    // Trivial classifier
    Classifier* tc = new TrivialClassifier(LEGITIMATE);
    // SVM classifier
    Classifier* svm = new SVMClassifier(svmSoftMargin == 1);
    // Perceptron
    Classifier* perceptron = new PerceptronClassifier();
    // SVM soft-margin
    Classifier* svmSoft = new SVMClassifier(true);
    // 1 of 2 (SVMSoft/NB)
    Classifier* svmnb = new OneOfTwoClassifier(svmSoft, nb, SPAM);
    // 2 of 3 (SVMSoft/NB/Perceptron);
    Classifier* svmnbperceptron = new TwoOfThreeClassifier(svmSoft, nb, perceptron);
    
    if (doTrivial)
        ClassifierPerformanceEstimator::testClassifier(tc, trainingSet, validationSet);
    if (doPerceptron)
        ClassifierPerformanceEstimator::testClassifier(perceptron, trainingSet, validationSet);    
    if (doNb)
        ClassifierPerformanceEstimator::testClassifier(nb, trainingSet, validationSet);
    if (doSvm)
        ClassifierPerformanceEstimator::testClassifier(svm, trainingSet, validationSet);
    if (doKnn)
        ClassifierPerformanceEstimator::testClassifier(knn, trainingSet, validationSet);
    if (doSvmNb)
        ClassifierPerformanceEstimator::testClassifier(svmnb, trainingSet, validationSet);
    if (doSvmNbPerceptron)
        ClassifierPerformanceEstimator::testClassifier(svmnbperceptron, trainingSet, validationSet);
    
  
    cout << "Completed" << endl;
  }
  catch(Exception e){
    cerr << "Error: " << e.message << endl;
    return 1;
  }
  return 0;
}
